/**   @file  oomfill.c
 *   @brief  Library safely (rlimited) consumming RAM to trigger OOMs (only on POSIX systems)
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  21/06/2024
 *      Revision:  none
 * Last modified:  2024-12-11 12:28
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "liboom/oomfill.h"                         /* OOM simulation */

#include <assert.h>
#include <stdio.h>                              /* printf */
#include <stdlib.h>                             /* abort() */
#include <errno.h>                              /* errno */
#include <sys/resource.h>                       /* setrlimit/getrlimit */
#include <string.h>                             /* strerror() */
#include <alloca.h>                             /* alloca() */
#include <unistd.h>                             /* sysconf() */

/** Maximum number of fragmented blocks to allocate
 *
 * I never reached more than 350 on my systems.
 * The value needs to be less than UINT_MAX
 * */
#define RAMBLOCKS_MAX 1000

/** RAM eating pointer
 *
 * Used to store the huge memory block used to fill the available memory */
static void* _oomblocks[RAMBLOCKS_MAX] = {0};

/** getrlimit helper with abort on fail */
static int checked_getrlimit(int resource, struct rlimit *rlim) {
    /* Get current limit values */
    if (getrlimit(resource, rlim) != 0) {
        /* Can occur, thus not ignored, but impossible to trigger for gcov/lcov */
        fprintf (stderr,"%s:%d getrlimit() failed with errno=%d %s\n",
                 __FILE__,__LINE__, errno,strerror(errno));
        abort();
    }
    return 0;
}

/** Disabled fill function
 *
 * Used as oomfill_fill when disabled. Thus, the same test can be surrounded by
 * fill/free, but the actual RAM pressure will or will not be activated.
 *
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
static size_t oomfill_fill_preinit(const size_t minHeap, const size_t minStack) {
    (void)minHeap;
    (void)minStack;
    assert(NULL==_oomblocks[0]);
    return 0;
}

/** Starts an almost OOM single and simple test
 *
 * This function is only enabled after a call to oomfill_enable and will have no
 * effect otherwise. Its goal is to completely fill the RAM until the very last
 * bytes, to keep only between minHeap and maxHeap bytes available in the heap
 * and minStack bytes in the stack.
 *
 * If oomfill_enable was not invoked beforehand, this function will simply
 * return without any RAM consumption.
 *
 * It should be called immediately before the test and should be reverted with
 * oomfill_free immediately after. It can create so much pressure on the
 * available memory that a simple printf could fail.
 *
 * It is designed to fail if invoked twice as this is very probably a mistake in
 * the test code. Should you need to change the RAM filling values, first call
 * oomfill_free and reapply oomfill_fill. This ensure a really wanted behavior
 * and not a mistake in your test code.
 *
 * \param [in] minHeap
 * \param [in] minStack
 *
 * \return Returns the allocated size to fill the memory
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
size_t (*oomfill_fill)(const size_t minHeap, const size_t minStack)=oomfill_fill_preinit;

/** Disabled free function
 *
 * Used as oomfill_free when disabled. Thus, the same test can be surrounded by
 * fill/free, but the actual RAM pressure will or will not be activated.
 *
 * \sa oomfill_free
 * \sa oomfill_fill
 * */
static void oomfill_free_preinit() {
    assert(NULL==_oomblocks[0]);
}

/** Ends a single simple OOM test
 *
 * If this function is invoked between an oomfill_enable and an oomfill_disable
 * invocations, it deallocates the RAM allocated by the oomfill_fill function. It
 * should be called immediately after the single and simple test because the RAM
 * pressure can even make a printf to fail.
 *
 * If called without a prior oomfill_enable invocation, this function simply
 * returns without any action.
 *
 * This function is designed to fail and abort the process if invocated whereas
 * there is no current RAM allocated, when called twice, for example. This helps
 * to avoid mistakes in the test scenario.
 *
 * */
void (*oomfill_free)()=oomfill_free_preinit;

/** Disabled oomfill_enable function
 *
 * Used as oomfill_enable before oomfill_config was invoked, which should never
 * happen and will fail to detect mistakes in the test code.
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * */
static size_t oomfill_enable_preinit(const size_t softlimit) {
    (void)softlimit;

    /* Should not be called without configuration */
    fprintf(stderr,"%s:%d oomfill_enable called without oomfill_config before\n",__FILE__,__LINE__);
    abort();
}

/** Starts a new oomfill environment or reconfigure the soft limit
 *
 * This function can only be invoked after at least a first call to
 * oomfill_config to initialize the oomfill helpers environment.  If invocated
 * before, it will abort the current process to help detecting mistakes in the
 * test code.
 *
 * It will (soft)limit the current process and his children to the provided
 * value in bytes. Then, it will enable the fill/free oomfill helper functions
 * which are disabled otherwise.
 *
 * It is designed to be invoked at the begining of a test case to apply for all
 * tests in this testcase.
 *
 * In case of any unexpected behavior, it should abort the current process.
 *
 * \param [in] softlimit Soft limit to set in bytes. It has to be less than the
 * hardlimit otherwise it will fail and abort. If set to 0, it will apply the
 * same value as the hardlimit.
 *
 * \return Returns the actually set value as a softlimit, either the requested
 * value or the hardlimit in case of 0 requested.
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
size_t (*oomfill_enable)(const size_t softlimit)=oomfill_enable_preinit;

/** Disabled oomfill_disable function
 *
 * Used as oomfill_disable before oomfill_config was invoked, which should never
 * happen and will fail to detect mistakes in the test code.
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * */
static size_t oomfill_disable_preinit() {
    /* Should not be called before configuration */
    fprintf(stderr,"%s:%d oomfill_disable called without oomfill_config before\n",__FILE__,__LINE__);
    abort();
}

/** Stops the current oomfill
 *
 * This function can be invoked any time after the oomfill_config initialized
 * the environment. If invocated before, it will abort the current process to
 * help detecting mistakes in the test code.
 *
 * This function disables the fill/free oomfill functions and free the allocated
 * RAM before reverting the soft limit to the hard limit value.
 *
 * It is designed to be invoked t the end of a test case.
 *
 * Despite it could be invoked twice or without a prior call to
 * oomfill_enable, it is illegal and not allowed to help detecting
 * mistakes in the test code.
 *
 * \return Returns the applied soft limit, which should be the same as the hard
 * limit.
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
size_t (*oomfill_disable)()=oomfill_disable_preinit;

/** Finds biggest available RAM block and allocate it
 *
 * This functions tries to find and allocate the biggest available memory block
 * by dichotomy (Olog2 complexity) between 0 and rlim_cur soft limit. It returns
 * the pointer.
 *
 * \param [in,out] The allocated pointer if successfull, NULL otherwise
 *
 * \return The size of the allocated block, 0 if not block was allocated.
 *
 * */
static size_t oomfill_getbiggestblock(void** p_ramblock) {
    /* This function could receive an optimized "max" value and return the final
     * "max" value to the caller. So, it could be used as the next invocation
     * starting "max" value. This would save few loop iterations, at the cost of
     * extra stack usage, I chose to not pass this value as a parameter to avoid
     * consuming stack.
     * The function starts to search between 0..rlim_cur which has very little
     * impact given the Olog2 complexity. It could be optimized and start to
     * search between 0..sysconf(AVPHYSPAGE*PAGESIZE), assuming that sysconf is
     * faster than few loop iterations. */

    /* Use static to avoid stack allocation/free */
    static size_t max,cur;
    static struct rlimit limit;

    assert(NULL!=p_ramblock);
    assert(NULL==*p_ramblock);

    /* Get the current limits */
    checked_getrlimit(RLIMIT_AS, &limit);
    max = limit.rlim_cur;

    /* Restart the whole process if cur can not be allocated at the end */
    while ((max>0)&&(NULL==*p_ramblock)) {
        static size_t min;
        /* Iterate quickly (Olog2) to converge to the biggest available RAM block */
        min = 0;
        while (max>min) {
            cur = min+(max-min)/2; /* To avoid overflow */
            if (NULL==(*p_ramblock = malloc(cur))) {
                max = cur;
            } else {
                min = cur+1;
                free(*p_ramblock);
            }
        }
        cur -= 1;
        *p_ramblock=malloc(cur);
    }

    return cur;
}

/** Enabled fill function
 *
 * Used as oomfill_fill when enabled.
 *
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
static size_t oomfill_fill_postinit(const size_t minHeap, const size_t minStack) {
    unsigned int l_numblock;
    size_t l_sum;
    void* volatile l_reservedheap;
    void* l_reservedstack;

    /* Probably a mistake in the test code, do not accept despite we could */
    if (NULL!=_oomblocks[0]) {
        /* Dirty hack to free some RAM and allow abort to SIGABRT */
        free(_oomblocks[0]);
        fprintf (stderr,"%s:%d oomblocks are already allocated\n", __FILE__,__LINE__);
        abort();
    }

    /* Reserve/Protect stack bytes which will be auto freed at return */
    /* A failure means a stackoverflow, which is not recoverable and will abort
     * the process anyway. */
    if (0<minStack)
        l_reservedstack=alloca(minStack);
    (void)l_reservedstack;

    /* Reserve/Protect heap bytes which will be released before return */
    l_reservedheap=NULL;
    if (0<minHeap)
        if (NULL==(l_reservedheap=malloc(minHeap))) {
            /* This will fail if minHeap is higher than rlimit */
            fprintf(stderr,"%s:%d Failed to reserve minheap bytes\n",__FILE__,__LINE__);
            abort();
        }

    /* Find and allocate the biggest available RAM blocks until no mre RAM
     * available or all _oomblocks are allocated */
    l_numblock=0;
    l_sum = 0;
    l_sum += oomfill_getbiggestblock(&(_oomblocks[l_numblock++]));
    while ((NULL!=_oomblocks[l_numblock-1])&&((RAMBLOCKS_MAX-1)>l_numblock))
        l_sum += oomfill_getbiggestblock(&(_oomblocks[l_numblock++]));
    /* Either already NULL, which stopped the while loop or reached the last
     * slot in the table, which stopped the while loop and the slot has to be
     * set to NULL */
    _oomblocks[l_numblock]=NULL;

    /* There can be less than 4 bytes available at this point !!! */

    /* Make the protected heap bytes available again */
    if (NULL!=l_reservedheap)
        free(l_reservedheap);

    return l_sum;
}

/** Enabled free function
 *
 * Used as oomfill_free when enabled
 *
 * \sa oomfill_free
 * \sa oomfill_fill
 * */
static void oomfill_free_postinit() {
    unsigned int l_i;

    /* Despite we could manage, abort to help detecting mistakes in test code */
    if (NULL==_oomblocks[0]) {
        fprintf(stderr,"%s:%d no blocks to free in oomfill_free.\n",__FILE__,__LINE__);
        abort();
    }

    /* Actually free allocated blocks and set their pointer to NULL */
    l_i = 0;
    while ((l_i<(RAMBLOCKS_MAX-1))&&(_oomblocks[l_i])) {
        free(_oomblocks[l_i]);
        _oomblocks[l_i] = NULL;
        l_i+=1;
    }
}

/** Enabled oomfill_enable function
 *
 * Used as oomfill_enable after oomfill_config was invoked
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * */
static size_t oomfill_enable_postinit(const size_t softlimit) {
    struct rlimit limit;
    size_t l_softlimit = softlimit;

    /* Probably a bug in oomfill */
    assert(NULL==_oomblocks[0]);

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);

    /* Check the requested value */
    if (0==l_softlimit) {
        /* Defaults to available physical RAM if requested 0 */
        l_softlimit = limit.rlim_max;
    };

    /* Soft limit available RAM */
    limit.rlim_cur = l_softlimit;

    /* Abort if setrlimit fails to avoid RAM bombing */
    if (setrlimit(RLIMIT_AS, &limit) != 0) {
        fprintf (stderr,"%s:%d setrlimit(cur=%lu, max=%lu) with errno=%d %s\n",
                 __FILE__,__LINE__,
                 (unsigned long)limit.rlim_cur, (unsigned long)limit.rlim_max,
                 errno,strerror(errno));
        abort();
    }

    /* Activate fill/free functions */
    oomfill_fill = oomfill_fill_postinit;
    oomfill_free = oomfill_free_postinit;

    /* Return the actual current soft limit */
    return limit.rlim_cur;
}

/** Enabled oomfill_disable function
 *
 * Used as oomfill_disable after oomfill_config was invoked
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * */
static size_t oomfill_disable_postinit() {
    struct rlimit limit;

    if (NULL!=_oomblocks[0]) {
        fprintf(stderr,"%s:%d RAM still allocated while calling oomfill_disable\n",__FILE__,__LINE__);
        abort();
    }

    /* Do not allow calling disable if not enabled to detect test mistakes */
    if ((oomfill_fill!=oomfill_fill_postinit)||(oomfill_free!=oomfill_free_postinit)) {
        fprintf(stderr,"%s:%d Impossible to disable oomfill if not previously enabled\n",__FILE__,__LINE__);
        abort();
    }

    /* Reset the soft limit to hard limit */
    oomfill_enable_postinit(0);

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);

    /* Restore disabled functors */
    oomfill_fill = oomfill_fill_preinit;
    oomfill_free = oomfill_free_preinit;

    /* Return the actual current soft limit, which is equal to hard limit */
    return limit.rlim_cur;
}

/** Sets the oomfill helpers hard rlimit and enables the oomfill helper features
 *
 * This function needs to be invoqued BEFORE any other helper from the
 * framework. It will configure an hard RAM limit for the current process and
 * each of his children. Then, it will enable the other oomfill helper
 * functions.
 *
 * It refuses to set a limit over the actually installed physical RAM to limit
 * pushing RAM pages to the swap. If the requested value is zero, it will
 * default to the actually installed physical RAM.
 *
 * If anything goes wrong or did not behaves as expected, it abort the process
 * to avoid RAM bombing and swapping.
 *
 * Despite it was designed to be invoked only once, from the main parent
 * process, it can be invoked several times as long as the hardlimit parameter
 * is always less than the previous call, otherwise it will fail and abort the
 * process.
 *
 * \param [in] hardlimit the size in bytes to limit the process to.
 *
 * \return Returns the actual configured size. It should be the same
 * value as the hardlimit parameter, otherwise something went wrong, it was not
 * detected and the process was not aborted (which is a bug to report)
 *
 * \sa oomfill_enable
 * \sa oomfill_disable
 * \sa oomfill_fill
 * \sa oomfill_free
 * */
size_t oomfill_config(const size_t hardlimit) {
    struct rlimit limit;
    size_t l_avail;
    size_t l_hardlimit = hardlimit;

    /* Probably a test implementation mistake */
    if (NULL!=_oomblocks[0]) {
        fprintf(stderr,"%s:%d Calling oomfill_config with allocated RAM blocks is not allowed.\n",
                __FILE__,__LINE__);
        abort();
    }

    /* Find *installed* physical RAM, 0 in case of failure */
    l_avail = (sysconf(_SC_PHYS_PAGES) * sysconf(_SC_PAGESIZE));

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);
    if (0==l_hardlimit) {
        /* Defaults to available physical RAM or already set rlimit
         * if 0 requested */
        l_hardlimit=(limit.rlim_max<l_avail?limit.rlim_max:l_avail);
    } else if (l_hardlimit>l_avail) {
        /* Fails if request is over available physical RAM to avoir swapping */
        fprintf(stderr,"%s:%d Requesing a limit %lu bigger than *installed* RAM %lu is not allowed.\n",
                __FILE__,__LINE__,
                (unsigned long)l_hardlimit,(unsigned long)l_avail);
        abort();
    }

    /* Hard limit available RAM to hardlimit globally with no way back */
    limit.rlim_cur = l_hardlimit;
    limit.rlim_max = l_hardlimit;

    /* Abort if setrlimit fails to avoid RAM bombing */
    if (setrlimit(RLIMIT_AS, &limit) != 0) {
        fprintf (stderr,"%s:%d setrlimit(cur=%lu, max=%lu) with errno=%d %s\n",
                 __FILE__,__LINE__, (unsigned long)limit.rlim_cur,
                 (unsigned long)limit.rlim_max, errno,strerror(errno));
        /* Get current limit values */
        checked_getrlimit(RLIMIT_AS, &limit);
        fprintf (stderr,"%s:%d getrlimit() is cur=%lu, max=%lu\n",
                 __FILE__,__LINE__, (unsigned long)limit.rlim_cur,
                 (unsigned long)limit.rlim_max);
        abort();
    }

    /* Activate enable/disable functions */
    oomfill_enable = oomfill_enable_postinit;
    oomfill_disable = oomfill_disable_postinit;

    /* Return the configured limit hard=soft */
    return limit.rlim_max;
}

bool oomfill_enabled() {
    return (oomfill_fill==oomfill_fill_postinit?true:false);
}
/* vim: set tw=80: */
